@@ -1,6 +1,6 @@
 """Kernel Tuner interface module
 
-This module contains the main functions that the Kernel Tuner
+This module contains the main functions that Kernel Tuner
 offers to its users.
 
 Author
@@ -38,33 +38,38 @@ import kernel_tuner.util as util
 import kernel_tuner.core as core
 
 from kernel_tuner.runners.sequential import SequentialRunner
+from kernel_tuner.runners.simulation import SimulationRunner
 
 from kernel_tuner.strategies import brute_force, random_sample, diff_evo, minimize, basinhopping, genetic_algorithm, pso, simulated_annealing, firefly_algorithm, bayes_opt
 
-strategy_map = {"brute_force": brute_force,
-                "random_sample": random_sample,
-                "minimize": minimize,
-                "basinhopping": basinhopping,
-                "diff_evo": diff_evo,
-                "genetic_algorithm": genetic_algorithm,
-                "pso": pso,
-                "simulated_annealing": simulated_annealing,
-                "firefly_algorithm": firefly_algorithm,
-                "bayes_opt": bayes_opt}
+strategy_map = {
+    "brute_force": brute_force,
+    "random_sample": random_sample,
+    "minimize": minimize,
+    "basinhopping": basinhopping,
+    "diff_evo": diff_evo,
+    "genetic_algorithm": genetic_algorithm,
+    "pso": pso,
+    "simulated_annealing": simulated_annealing,
+    "firefly_algorithm": firefly_algorithm,
+    "bayes_opt": bayes_opt
+}
+
 
 class Options(OrderedDict):
     """read-only class for passing options around"""
+
     def __getattr__(self, name):
         if not name.startswith('_'):
             return self[name]
         return super(Options, self).__getattr__(name)
+
     def __deepcopy__(self, _):
         return self
 
 
-_kernel_options = Options([
-    ("kernel_name", ("""The name of the kernel in the code.""", "string")),
-    ("kernel_source", ("""The CUDA, OpenCL, or C kernel code.
+_kernel_options = Options([("kernel_name", ("""The name of the kernel in the code.""", "string")),
+                           ("kernel_source", ("""The CUDA, OpenCL, or C kernel code.
             It is allowed for the code to be passed as a string, a filename, a function
             that returns a string of code, or a list when the code needs auxilliary files.
 
@@ -81,31 +86,46 @@ _kernel_options = Options([
             which will be used to pass a dict containing the parameters.
             The function should return a string with the source code for
             the kernel.""", "string or list and/or callable")),
-    ("lang", ("""Specifies the language used for GPU kernels. The kernel_tuner
+                           ("lang", ("""Specifies the language used for GPU kernels. The kernel_tuner
         automatically detects the language, but if it fails, you may specify
-        the language using this argument, currently supported: "CUDA",
+        the language using this argument, currently supported: "CUDA", "Cupy",
         "OpenCL", or "C".""", "string")),
-    ("problem_size", ("""An int or string, or 1,2,3-dimensional tuple
-            containing the size from which the grid dimensions of the kernel
-            will be computed.
+                           ("problem_size", ("""The size of the domain from which the grid dimensions
+            of the kernel are computed.
 
-            Do not divide the problem_size yourself by the thread block sizes.
-            The Kernel Tuner does this for you based on tunable parameters,
+            This can be specified using an int, string, function, or
+            1,2,3-dimensional tuple.
+
+            In general, do not divide the problem_size yourself by the thread block sizes.
+            Kernel Tuner does this for you based on tunable parameters,
             called "block_size_x", "block_size_y", and "block_size_z".
             If more or different parameters divide the grid dimensions use
             grid_div_x/y/z options to specify this.
 
-            You are allowed to use a string to specify the problem
-            size. Within a string you are allowed to write Python
+            In most use-cases the problem_size is specified using a single integer
+            or a tuple of integers,
+            but Kernel Tuner supports more advanced use cases where the problem_size
+            itself depends on the tunable parameters in some way.
+
+            You are allowed to use a function or string to specify the problem_size.
+            A function should accept a dictionary with the tunable parameters
+            for this kernel configuration and directly return a tuple
+            that specifies the problem size in all dimensions.
+
+            When passing a string, you are allowed to write Python
             arithmetic and use the names of tunable parameters as variables
-            in these expressions.
-            The Kernel Tuner will replace instances of the tunable parameters
-            with their current value when computing the grid dimensions.
-            See the reduction CUDA example for an example use of this feature.""",
-            "string, int, or tuple(int or string, ..)")),
-    ("arguments", ("""A list of kernel arguments, use numpy arrays for
+            in these expressions. Kernel Tuner will replace instances of the tunable
+            parameters with their current value when computing the grid dimensions.
+            This option exists for convenience, but do note that using a lambda
+            function is probably safer. The string notation should only return
+            the problem size for one dimension, but can be used inside
+            a tuple, possibly in combination with integers or more strings in
+            different dimensions.
+
+            See the reduction CUDA example for an example use of this feature.""", "callable, string, int, or tuple(int or string, ..)")),
+                           ("arguments", ("""A list of kernel arguments, use numpy arrays for
             arrays, use numpy.int32 or numpy.float32 for scalars.""", "list")),
-    ("grid_div_x", ("""A list of names of the parameters whose values divide
+                           ("grid_div_x", ("""A list of names of the parameters whose values divide
             the grid dimensions in the x-direction.
             The product of all grid divisor expressions is computed before dividing
             the problem_size in that dimension. Also note that the divison is treated
@@ -116,58 +136,54 @@ _kernel_options = Options([
             used if necessary inside the string containing a parameter name. For
             example, in some cases you may want to divide the problem size in the
             x-dimension with the number of warps rather than the number of threads
-            in a block, in such cases one could use ["block_size_x/32"].
+            in a block, in such cases one could for example use ["block_size_x/32"].
+            Another option is to pass a function to grid_div_x that accepts a
+            dictionary with the tunable parameters and returns the grid divisor
+            in this dimension, for example: grid_div_x=lambda p:p["block_size_x"]/32.
 
             If not supplied, ["block_size_x"] will be used by default, if you do not
-            want any grid x-dimension divisors pass an empty list.""", "list")),
-    ("grid_div_y", ("""A list of names of the parameters whose values divide
+            want any grid x-dimension divisors pass an empty list.""", "callable or list")),
+                           ("grid_div_y", ("""A list of names of the parameters whose values divide
             the grid dimensions in the y-direction, ["block_size_y"] by default.
             If you do not want to divide the problem_size, you should pass an empty list.
             See grid_div_x for more details.""", "list")),
-    ("grid_div_z", ("""A list of names of the parameters whose values divide
+                           ("grid_div_z", ("""A list of names of the parameters whose values divide
             the grid dimensions in the z-direction, ["block_size_z"] by default.
             If you do not want to divide the problem_size, you should pass an empty list.
             See grid_div_x for more details.""", "list")),
-    ("smem_args", ("""CUDA-specific feature for specifying shared memory options
+                           ("smem_args", ("""CUDA-specific feature for specifying shared memory options
             to the kernel. At the moment only 'size' is supported, but setting the
             shared memory configuration on Kepler GPUs for example could be added
             in the future. Size should denote the number of bytes for to use when
-            dynamically allocating shared memory.""",
-            "dict(string: numpy object)")),
-    ("cmem_args", ("""CUDA-specific feature for specifying constant memory
+            dynamically allocating shared memory.""", "dict(string: numpy object)")),
+                           ("cmem_args", ("""CUDA-specific feature for specifying constant memory
             arguments to the kernel. In OpenCL these are handled as normal
             kernel arguments, but in CUDA you can copy to a symbol. The way you
             specify constant memory arguments is by passing a dictionary with
             strings containing the constant memory symbol name together with numpy
-            objects in the same way as normal kernel arguments.""",
-            "dict(string: numpy object)")),
-    ("texmem_args", ("""CUDA-specific feature for specifying texture memory
+            objects in the same way as normal kernel arguments.""", "dict(string: numpy object)")),
+                           ("texmem_args", ("""CUDA-specific feature for specifying texture memory
             arguments to the kernel. You specify texture memory arguments by passing a
             dictionary with strings containing the texture reference name together with
             the texture contents. These contents can be either simply a numpy object,
             or a dictionary containing the numpy object under the key 'array' plus the
             configuration options 'filter_mode' ('point' or 'linear), 'address_mode'
             (a list of 'border', 'clamp', 'mirror', 'wrap' per axis),
-            'normalized_coordinates' (True/False).""",
-            "dict(string: numpy object or dict)")),
-    ("block_size_names", ("""A list of strings that replace the defaults for the names
+            'normalized_coordinates' (True/False).""", "dict(string: numpy object or dict)")),
+                           ("block_size_names", ("""A list of strings that replace the defaults for the names
             that denote the thread block dimensions. If not passed, the behavior
-            defaults to ``["block_size_x", "block_size_y", "block_size_z"]``""",
-            "list(string)"))
-    ])
+            defaults to ``["block_size_x", "block_size_y", "block_size_z"]``""", "list(string)"))])
 
-
-_tuning_options = Options([
-    ("tune_params", ("""A dictionary containing the parameter names as keys,
+_tuning_options = Options([("tune_params", ("""A dictionary containing the parameter names as keys,
             and lists of possible parameter settings as values.
-            The Kernel Tuner will try to compile and benchmark all possible
+            Kernel Tuner will try to compile and benchmark all possible
             combinations of all possible values for all tuning parameters.
             This typically results in a rather large search space of all
             possible kernel configurations.
 
             For each kernel configuration, each tuning parameter is
             replaced at compile-time with its current value.
-            Currently, the Kernel Tuner uses the convention that the following
+            Currently, Kernel Tuner uses the convention that the following
             list of tuning parameters are used as thread block dimensions:
 
                 * "block_size_x"   thread block (work group) x-dimension
@@ -177,27 +193,31 @@ _tuning_options = Options([
             Options for changing these defaults may be added later. If you
             don't want the thread block dimensions to be compiled in, you
             may use the built-in variables blockDim.xyz in CUDA or the
-            built-in function get_local_size() in OpenCL instead.""",
-            "dict( string : [...]")),
-    ("restrictions", ("""A list of strings containing boolean expression that
-        limit the search space in that they must be satisfied by the kernel
-        configuration. These expressions must be true for the configuration
+            built-in function get_local_size() in OpenCL instead.""", "dict( string : [...]")),
+                           ("restrictions", ("""An option to limit the search space with restrictions.
+        The restrictions can be specified using a function or a list of strings.
+        The function should take one argument, namely a dictionary with the
+        tunable parameters of the kernel configuration, if the function returns
+        True the configuration is considered to be part of the search space, or
+        False otherwise.
+        The other way to specify restrictions is using a list of strings
+        containing boolean expression that must be satisfied by the kernel
+        configuration. These expressions must all be true for the configuration
         to be part of the search space. For example:
         restrictions=["block_size_x==block_size_y*tile_size_y"] limits the
         search to configurations where the block_size_x equals the product
         of block_size_y and tile_size_y.
-        The default is None.""", "list")),
-    ("answer", ("""A list of arguments, similar to what you pass to arguments,
+        The default is None.""", "callable or list(strings)")),
+                           ("answer", ("""A list of arguments, similar to what you pass to arguments,
         that contains the expected output of the kernel after it has executed
         and contains None for each argument that is input-only. The expected
         output of the kernel will then be used to verify the correctness of
-        each kernel in the parameter space before it will be benchmarked.""",
-        "list")),
-    ("atol", ("""The maximum allowed absolute difference between two elements
+        each kernel in the parameter space before it will be benchmarked.""", "list")),
+                           ("atol", ("""The maximum allowed absolute difference between two elements
         in the output and the reference answer, as passed to numpy.allclose().
         Ignored if you have not passed a reference answer. Default value is
         1e-6, that is 0.000001.""", "float")),
-    ("verify", ("""Python function used for output verification. By default,
+                           ("verify", ("""Python function used for output verification. By default,
         numpy.allclose is used for output verification, if this does not suit
         your application, you can pass a different function here.
 
@@ -209,7 +229,7 @@ _tuning_options = Options([
         passed that was specified using the atol option to tune_kernel.
         The function should return True when the output passes the test, and
         False when the output fails the test.""", "func(ref, ans, atol=None)")),
-    ("strategy", ("""Specify the strategy to use for searching through the
+                           ("strategy", ("""Specify the strategy to use for searching through the
         parameter space, choose from:
 
             * "brute_force" (default),
@@ -247,7 +267,7 @@ _tuning_options = Options([
         "bayes_opt" uses Bayesian Optimization.
 
         """, "")),
-    ("strategy_options", ("""A dict with options for the tuning strategy
+                           ("strategy_options", ("""A dict with options for the tuning strategy
 
         Example usage:
 
@@ -303,10 +323,9 @@ _tuning_options = Options([
         the maximum number of function evaluations.
 
     """, "dict")),
-    ("iterations", ("""The number of times a kernel should be executed and
-        its execution time measured when benchmarking a kernel, 7 by default.""",
-        "int")),
-    ("verbose", ("""Sets whether or not to report about configurations that
+                           ("iterations", ("""The number of times a kernel should be executed and
+        its execution time measured when benchmarking a kernel, 7 by default.""", "int")),
+                           ("verbose", ("""Sets whether or not to report about configurations that
         were skipped during the search. This could be due to several reasons:
 
             * kernel configuration fails one or more restrictions
@@ -315,37 +334,35 @@ _tuning_options = Options([
             * too many resources requested for launch
 
         verbose is False by default.""", "bool")),
-    ("cache", ("""filename for caching/logging benchmarked instances
+                           ("cache", ("""filename for caching/logging benchmarked instances
         filename uses suffix ".json"
         if the file exists it is read and tuning continues from this file
-        """, "string"))
-    ])
+        """, "string")), ("metrics", ("specifies user-defined metrics", "OrderedDict")),
+                           ("simulation_mode", ("Simulate an auto-tuning search from an existing cachefile", "bool")),
+                           ("observers", ("""A list of BenchmarkObservers""", "list"))])
 
-_device_options = Options([
-    ("device", ("""CUDA/OpenCL device to use, in case you have multiple
+_device_options = Options([("device", ("""CUDA/OpenCL device to use, in case you have multiple
         CUDA-capable GPUs or OpenCL devices you may use this to select one,
         0 by default. Ignored if you are tuning host code by passing
         lang="C".""", "int")),
-    ("platform", ("""OpenCL platform to use, in case you have multiple
+                           ("platform", ("""OpenCL platform to use, in case you have multiple
         OpenCL platforms you may use this to select one,
         0 by default. Ignored if not using OpenCL. """, "int")),
-    ("quiet", ("""Control whether or not to print to the console which
+                           ("quiet", ("""Control whether or not to print to the console which
         device is being used, False by default""", "boolean")),
-    ("compiler", ("""A string containing your preferred compiler,
-        only effective with lang="C". """, "string")),
-    ("compiler_options", ("""A list of strings that specify compiler
-        options.""", "list(string)"))
-    ])
-
+                           ("compiler", ("""A string containing your preferred compiler,
+        only effective with lang="C". """, "string")), ("compiler_options", ("""A list of strings that specify compiler
+        options.""", "list(string)"))])
 
 
 def _get_docstring(opts):
     docstr = ""
     for k, v in opts.items():
         docstr += "    :param " + k + ": " + v[0] + "\n"
-        docstr += "    :type "  + k + ": " + v[1] + "\n\n"
+        docstr += "    :type " + k + ": " + v[1] + "\n\n"
     return docstr
 
+
 _tune_kernel_docstring = """ Tune a CUDA kernel given a set of tunable parameters
 
 %s
@@ -360,18 +377,16 @@ _tune_kernel_docstring = """ Tune a CUDA kernel given a set of tunable parameter
 
 #"""
 
-def tune_kernel(kernel_name, kernel_string, problem_size, arguments,
-                tune_params, grid_div_x=None, grid_div_y=None, grid_div_z=None,
-                restrictions=None, answer=None, atol=1e-6, verify=None, verbose=False,
-                lang=None, device=0, platform=0, smem_args=None, cmem_args=None, texmem_args=None,
-                compiler=None, compiler_options=None, log=None,
-                iterations=7, block_size_names=None, quiet=False, strategy=None, strategy_options=None,
-                cache=None):
+
+def tune_kernel(kernel_name, kernel_string, problem_size, arguments, tune_params, grid_div_x=None, grid_div_y=None, grid_div_z=None, restrictions=None,
+                answer=None, atol=1e-6, verify=None, verbose=False, lang=None, device=0, platform=0, smem_args=None, cmem_args=None, texmem_args=None,
+                compiler=None, compiler_options=None, log=None, iterations=7, block_size_names=None, quiet=False, strategy=None, strategy_options=None,
+                cache=None, metrics=None, simulation_mode=False, observers=None):
 
     if log:
         logging.basicConfig(filename=kernel_name + datetime.now().strftime('%Y%m%d-%H:%M:%S') + '.log', level=log)
 
-    kernel_source = core.KernelSource(kernel_string, lang)
+    kernel_source = core.KernelSource(kernel_name, kernel_string, lang)
 
     _check_user_input(kernel_name, kernel_source, arguments, block_size_names)
 
@@ -425,46 +440,45 @@ def tune_kernel(kernel_name, kernel_string, problem_size, arguments,
     else:
         strategy = brute_force
 
+    # select the runner for this job based on input
+    SelectedRunner = SimulationRunner if simulation_mode else SequentialRunner
+    with SelectedRunner(kernel_source, kernel_options, device_options, iterations, observers) as runner:
 
-    runner = SequentialRunner(kernel_source, kernel_options, device_options, iterations)
+        #the user-specified function may or may not have an optional atol argument;
+        #we normalize it so that it always accepts atol.
+        tuning_options.verify = util.normalize_verify_function(tuning_options.verify)
 
-    #the user-specified function may or may not have an optional atol argument;
-    #we normalize it so that it always accepts atol.
-    tuning_options.verify = util.normalize_verify_function(tuning_options.verify)
+        #process cache
+        if cache:
+            if cache[-5:] != ".json":
+                cache += ".json"
 
-    #process cache
-    if cache:
-        if cache[-5:] != ".json":
-            cache += ".json"
-
-        util.process_cache(cache, kernel_options, tuning_options, runner)
-    else:
-        tuning_options.cache = {}
-        tuning_options.cachefile = None
-
-    #call the strategy to execute the tuning process
-    results, env = strategy.tune(runner, kernel_options, device_options, tuning_options)
-
-    #finished iterating over search space
-    if not device_options.quiet:
-        if results:     #checks if results is not empty
-            best_config = min(results, key=lambda x: x['time'])
-            units = getattr(runner, "units", None)
-            print("best performing configuration:", util.get_config_string(best_config, list(tune_params.keys()) + ['time'], units=units))
+            util.process_cache(cache, kernel_options, tuning_options, runner)
         else:
-            print("no results to report")
+            tuning_options.cache = {}
+            tuning_options.cachefile = None
 
-    if cache:
-        util.close_cache(cache)
+        #call the strategy to execute the tuning process
+        results, env = strategy.tune(runner, kernel_options, device_options, tuning_options)
 
-    del runner.dev
+        #finished iterating over search space
+        if not device_options.quiet:
+            if results:    #checks if results is not empty
+                best_config = min(results, key=lambda x: x['time'])
+                units = getattr(runner, "units", None)
+                print("best performing configuration:")
+                util.print_config_output(tune_params, best_config, device_options.quiet, metrics, units)
+            else:
+                print("no results to report")
+
+        if cache:
+            util.close_cache(cache)
 
     return results, env
 
 
 tune_kernel.__doc__ = _tune_kernel_docstring
 
-
 _run_kernel_docstring = """Compile and run a single kernel
 
     Compiles and runs a single kernel once, given a specific instance of the kernels tuning parameters.
@@ -479,7 +493,7 @@ _run_kernel_docstring = """Compile and run a single kernel
      * Execute the kernel on the GPU
      * Copy all data from the GPU back to the host and return it as a list of Numpy arrays
 
-    This function was added to the Kernel Tuner mostly to allow easy testing for kernel correctness.
+    This function was added to Kernel Tuner mostly to allow easy testing for kernel correctness.
     On purpose, the interface is a lot like `tune_kernel()`.
 
 %s
@@ -494,15 +508,13 @@ _run_kernel_docstring = """Compile and run a single kernel
 """ % _get_docstring(_kernel_options) + _get_docstring(_device_options)
 
 
-def run_kernel(kernel_name, kernel_string, problem_size, arguments,
-               params, grid_div_x=None, grid_div_y=None, grid_div_z=None,
-               lang=None, device=0, platform=0, smem_args=None, cmem_args=None, texmem_args=None, compiler=None, compiler_options=None,
-               block_size_names=None, quiet=False, log=None):
+def run_kernel(kernel_name, kernel_string, problem_size, arguments, params, grid_div_x=None, grid_div_y=None, grid_div_z=None, lang=None, device=0, platform=0,
+               smem_args=None, cmem_args=None, texmem_args=None, compiler=None, compiler_options=None, block_size_names=None, quiet=False, log=None):
 
     if log:
         logging.basicConfig(filename=kernel_name + datetime.now().strftime('%Y%m%d-%H:%M:%S') + '.log', level=log)
 
-    kernel_source = core.KernelSource(kernel_string, lang)
+    kernel_source = core.KernelSource(kernel_name, kernel_string, lang)
 
     _check_user_input(kernel_name, kernel_source, arguments, block_size_names)
 
@@ -512,62 +524,59 @@ def run_kernel(kernel_name, kernel_string, problem_size, arguments,
     device_options = Options([(k, opts[k]) for k in _device_options.keys()])
 
     #detect language and create the right device function interface
-    dev = core.DeviceInterface(kernel_source, iterations=1, **device_options)
+    with core.DeviceInterface(kernel_source, iterations=1, **device_options) as dev:
 
-    #move data to the GPU
-    gpu_args = dev.ready_argument_list(arguments)
+        #move data to the GPU
+        gpu_args = dev.ready_argument_list(arguments)
 
-    instance = None
-    try:
-        #create kernel instance
-        instance = dev.create_kernel_instance(kernel_source, kernel_options, params, False)
-        if instance is None:
-            raise Exception("cannot create kernel instance, too many threads per block")
+        instance = None
+        try:
+            #create kernel instance
+            instance = dev.create_kernel_instance(kernel_source, kernel_options, params, False)
+            if instance is None:
+                raise Exception("cannot create kernel instance, too many threads per block")
 
-        # see if the kernel arguments have correct type
-        util.check_argument_list(instance.name, instance.kernel_string, arguments)
+            # see if the kernel arguments have correct type
+            util.check_argument_list(instance.name, instance.kernel_string, arguments)
 
-        #compile the kernel
-        func = dev.compile_kernel(instance, False)
-        if func is None:
-            raise Exception("cannot compile kernel, too much shared memory used")
+            #compile the kernel
+            func = dev.compile_kernel(instance, False)
+            if func is None:
+                raise Exception("cannot compile kernel, too much shared memory used")
 
-        #add constant memory arguments to compiled module
-        if cmem_args is not None:
-            dev.copy_constant_memory_args(cmem_args)
-        #add texture memory arguments to compiled module
-        if texmem_args is not None:
-            dev.copy_texture_memory_args(texmem_args)
-    finally:
-        #delete temp files
-        if instance is not None:
-            instance.delete_temp_files()
+            #add constant memory arguments to compiled module
+            if cmem_args is not None:
+                dev.copy_constant_memory_args(cmem_args)
+            #add texture memory arguments to compiled module
+            if texmem_args is not None:
+                dev.copy_texture_memory_args(texmem_args)
+        finally:
+            #delete temp files
+            if instance is not None:
+                instance.delete_temp_files()
 
-    #run the kernel
-    if not dev.run_kernel(func, gpu_args, instance):
-        raise Exception("runtime error occured, too many resources requested")
+        #run the kernel
+        if not dev.run_kernel(func, gpu_args, instance):
+            raise Exception("runtime error occured, too many resources requested")
 
-    #copy data in GPU memory back to the host
-    results = []
-    for i, arg in enumerate(arguments):
-        if numpy.isscalar(arg):
-            results.append(arg)
-        else:
-            results.append(numpy.zeros_like(arg))
-            dev.memcpy_dtoh(results[-1], gpu_args[i])
-
-    #trying to make run_kernel work nicely with the Nvidia Visual Profiler
-    del dev
+        #copy data in GPU memory back to the host
+        results = []
+        for i, arg in enumerate(arguments):
+            if numpy.isscalar(arg):
+                results.append(arg)
+            else:
+                results.append(numpy.zeros_like(arg))
+                dev.memcpy_dtoh(results[-1], gpu_args[i])
 
     return results
 
 
 run_kernel.__doc__ = _run_kernel_docstring
 
+
 def _check_user_input(kernel_name, kernel_source, arguments, block_size_names):
     # see if the kernel arguments have correct type
     kernel_source.check_argument_lists(kernel_name, arguments)
 
     # check for types and length of block_size_names
     util.check_block_size_names(block_size_names)
-
